import os
import random
import requests
from omspy.base import Broker, pre, post
from omspy.utils import dict_filter
from requests_oauthlib import OAuth2Session

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from typing import Dict, List


def get_authorization_url():
    oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scope)
    authorization_url, _state = oauth.authorization_url(
        authorization_base_url, access_type="authorization_code"
    )
    return authorization_url


def fetch_all_contracts(exchanges=["NSE", "NFO"]):
    """
    Fetch all contracts for the given list of exchanges
    exchanges
        exchanges as a list
    """
    url = "https://masterswift.mastertrust.co.in/api/v2/contracts.json?exchanges={exc}"
    # All contracts are stored as dictionary keys
    contracts = {}
    for e in exchanges:
        url2 = url.format(exc=e)
        req = requests.get(url2).json()
        for k, v in req.items():
            for c in v:
                symbol = c["trading_symbol"]
                code = c["code"]
                contracts[f"{e}:{symbol}"] = code
    return contracts


def get_instrument_token(contracts, exchange, symbol):
    """
    Fetch the instrument token
    contracts
        the contracts master as a dictionary
    exchange
        exchange to look up for
    symbol
        symbol to look up for
    """
    return contracts.get(f"{exchange}:{symbol}")


class MasterTrust(Broker):
    """
    Automated Trading class
    """

    def __init__(
        self,
        client_id,
        password,
        PIN,
        secret,
        exchange="NSE",
        product="MIS",
        token_file="token.tok",
    ):
        self.filter = dict_filter
        self._client_id = client_id
        self._password = password
        self._pin = PIN
        self._secret = secret
        self.exchange = exchange
        self.product = product
        self._store_access_token = True
        self._access_token = None
        self.token_file = token_file
        self.base_url = "https://masterswift-beta.mastertrust.co.in"
        self.authorization_base_url = f"{self.base_url}/oauth2/auth"
        self.token_url = f"{self.base_url}/oauth2/token"
        super(MasterTrust, self).__init__()
        try:
            with open(self.token_file, "r") as f:
                access_token = f.read()
            self._access_token = access_token
        except Exception as e:
            print("Token not found", e)

        self._set_headers()
        self._sides = {"BUY": "SELL", "SELL": "BUY"}

    @property
    def headers(self):
        return self._headers

    @property
    def access_token(self):
        return self._access_token

    @property
    def client_id(self):
        return self._client_id

    def _set_headers(self):
        self._headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self._access_token}",
            "Cache-Control": "no-cache",
        }

    def get_authorization_url(
        self, client_id="APIUSER", redirect_uri="http://127.0.0.1/", scope=["orders"]
    ):
        oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scope)
        authorization_url, _state = oauth.authorization_url(
            self.authorization_base_url, access_type="authorization_code"
        )
        return authorization_url

    def get_access_token(self, url, redirect_uri="http://127.0.0.1/", scope=["orders"]):
        # to make oauth2 work with http
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
        oauth = OAuth2Session("APIUSER", redirect_uri=redirect_uri, scope=scope)
        token = oauth.fetch_token(
            self.token_url, authorization_response=url, client_secret=self._secret
        )
        access_token = token["access_token"]
        self._access_token = access_token
        with open(self.token_file, "w") as f:
            f.write(access_token)
        return access_token

    def _shortcuts(self):
        """
        Provides shortcuts to master trust function
        """
        pass

    def authenticate(self, force=False):
        """
        Authenticates a session if access token is already
        available by looking at the token.tok file.
        In case authentication fails, try a fresh login
        force
            Force an authentication even if tokens exists
        """
        try:
            if not (force):
                with open(self.token_file, "r") as f:
                    access_token = f.read()
                self._access_token = access_token
                self._set_headers()
            else:
                login_url = self._login()
                access_token = self.get_access_token(login_url)
                self._access_token = access_token
                self._set_headers()
        except Exception as e:
            print(e)
            login_url = self._login()
            access_token = self.get_access_token(login_url)
            self._access_token = access_token
            self._set_headers()

    def _login(self):
        import time

        options = Options()
        options.add_argument("--headless")
        options.add_argument("--disable-gpu")
        driver = webdriver.Chrome(options=options)
        url = self.get_authorization_url()
        driver.get(url)
        time.sleep(2)
        WebDriverWait(driver, 45).until(
            EC.presence_of_element_located((By.CLASS_NAME, "btn-container"))
        )
        driver.find_element_by_name("login_id").send_keys(self._client_id)
        driver.find_element_by_name("password").send_keys(self._password)
        driver.find_element_by_xpath('//button[@type="submit"]').click()
        time.sleep(2)
        WebDriverWait(driver, 45).until(
            EC.presence_of_element_located((By.CLASS_NAME, "btn-container"))
        )
        driver.find_element_by_xpath('//input[@type="password"]').send_keys(self._pin)
        driver.find_element_by_xpath('//button[@type="submit"]').click()
        time.sleep(2)
        current_url = driver.current_url
        driver.close()
        return current_url

    def _response(self, response, full=False):
        """
        response
            response is the raw response from broker
        full
            if True, the entire json response is returned
            useful for debugging purposes and getting extra information
        """
        try:
            resp = response.json()
            if full or (resp.get("status") == "error"):
                return resp
            else:
                return resp["data"]
        except:
            return {}

    def _get_instrument_token(self, symbol, exchange="NSE", contracts=None):
        if not (contracts):
            contracts = self.contracts
        return get_instrument_token(
            contracts=contracts, exchange=exchange, symbol=symbol
        )

    @property
    def profile(self):
        """
        Get the profile for the user
        """
        url = f"{self.base_url}/api/v1/user/profile"
        payload = {"client_id": self.client_id}
        resp = requests.get(url, headers=self.headers, params=payload)
        return self._response(resp)

    @post
    def positions(self):
        """
        Return only the positions for the day
        """
        url = f"{self.base_url}/api/v1/positions"
        payload = {"client_id": self.client_id, "type": "live"}
        resp = requests.get(url, headers=self.headers, params=payload)
        resp = self._response(resp)
        for r in resp:
            r["side"] = "BUY" if r.get("quantity", 0) > 0 else "SELL"
        return resp

    @post
    def completed_orders(self):
        """
        Return the completed orders for the day
        """
        url = f"{self.base_url}/api/v1/orders"
        payload = {"client_id": self.client_id, "type": "completed"}
        resp = requests.get(url, headers=self.headers, params=payload)
        return self._response(resp).get("orders", [])

    @post
    def pending_orders(self):
        """
        Return the completed orders for the day
        """
        url = f"{self.base_url}/api/v1/orders"
        payload = {"client_id": self.client_id, "type": "pending"}
        resp = requests.get(url, headers=self.headers, params=payload)
        return self._response(resp).get("orders", [])

    def orders(self):
        """
        Return the entire orderbook for the day including
        completed and pending orders
        """
        pending = self.pending_orders()
        completed = self.completed_orders()
        pending.extend(completed)
        return pending

    def trades(self):
        """
        Return the tradebook for the day
        """
        url = f"{self.base_url}/api/v1/trades"
        payload = {"client_id": self.client_id}
        resp = requests.get(url, headers=self.headers, params=payload)
        return self._response(resp).get("trades", [])

    def realized_mtm(self, positions=None):
        """
        Get the realized MTM
        """
        if not (positions):
            positions = self.positions()
        if len(positions) > 0:
            return sum([float(p["realized_mtm"]) for p in positions])
        else:
            # Return 0 in case of no transactions
            return 0

    def unrealized_mtm(self, positions=None):
        """
        Get the unrealized MTM
        """
        if not (positions):
            positions = self.positions()
        if len(positions) == 0:
            collect = {p["symbol"]: 0 for p in positions}
        else:
            collect = {}
            for p in positions:
                if p["quantity"] > 0:
                    collect[p["symbol"]] = (
                        p["ltp"] - (-p["net_amount"] / p["quantity"])
                    ) * p["quantity"] - p["realized_mtm"]
                elif p["quantity"] < 0:
                    collect[p["symbol"]] = (
                        p["ltp"] - (-p["net_amount"] / p["quantity"])
                    ) * p["quantity"] - p["realized_mtm"]
                else:
                    collect[p["symbol"]] = 0
        return sum(list(collect.values()))

    def mtm(self, positions=None):
        """
        Get the mtm
        """
        if not (positions):
            positions = self.positions()
        realized_mtm = self.realized_mtm(positions=positions)
        unrealized_mtm = self.unrealized_mtm(positions=positions)
        return realized_mtm + unrealized_mtm

    def net_qty(self, symbol):
        """
        Get the net quantity
        """
        positions = self.positions()
        if symbol is None:
            return {p["symbol"]: p["quantity"] for p in positions}
        else:
            for p in positions:
                if p["symbol"] == symbol:
                    return p["quantity"]
            return 0

    def order_place(self, **kwargs):
        """
        Place an order
        """
        order_args = dict(product="MIS", validity="DAY", exchange=self.exchange)
        url = f"{self.base_url}/api/v1/orders"
        symbol = kwargs.pop("symbol")
        side = kwargs.pop("side")
        exchange = kwargs.get("exchange", self.exchange)
        token = self._get_instrument_token(exchange=exchange, symbol=symbol)
        kwargs["instrument_token"] = token
        kwargs["order_side"] = side
        kwargs["client_id"] = self.client_id
        kwargs["user_order_id"] = random.randint(0, 1e9)
        order_args.update({"exchange": exchange})
        order_args.update(kwargs)
        payload = order_args.copy()
        resp = requests.post(url, headers=self.headers, params=payload)
        return self._response(resp)

    def order_modify(self, **kwargs):
        """
        Place an order
        """
        url = f"{self.base_url}/api/v1/orders"
        symbol = kwargs.pop("symbol")
        exchange = kwargs.get("exchange", self.exchange)
        token = self._get_instrument_token(exchange=exchange, symbol=symbol)
        kwargs["instrument_token"] = token
        kwargs["client_id"] = self.client_id
        payload = kwargs.copy()
        resp = requests.put(url, headers=self.headers, params=payload)
        return self._response(resp)

    def order_cancel(self, oms_order_id):
        """
        Place an order
        """
        url = f"{self.base_url}/api/v1/orders/{oms_order_id}"
        payload = {"client_id": self.client_id}
        resp = requests.delete(url, headers=self.headers, params=payload)
        return self._response(resp)

    def place_bracket_order(self, **kwargs):
        """
        Place a bracket order
        """
        url = f"{self.base_url}/api/v1/orders/bracket"
        symbol = kwargs.pop("symbol")
        side = kwargs.pop("side")
        exchange = kwargs.get("exchange", self.exchange)
        token = self._get_instrument_token(exchange=self.exchange, symbol=symbol)
        kwargs["instrument_token"] = token
        kwargs["order_side"] = side
        kwargs["client_id"] = self.client_id
        kwargs["user_order_id"] = 1000
        payload = kwargs.copy()
        resp = requests.post(url, headers=self.headers, params=payload)
        return self._response(resp)

    def exit_bracket_order(self, **kwargs):
        """
        Exit at existing bracket order
        """
        url = f"{self.base_url}/api/v1/orders/bracket/"
        payload = kwargs.copy()
        resp = requests.delete(url, headers=self.headers, params=payload)
        return resp.json()

    def modify_all_by_symbol(self, symbol, **kwargs):
        """
        Modify all pending orders for the given symbol
        symbol
            symbol for which orders to be changed
        kwargs
            provide modifications in the form of arguments
        Note
        ----
        1) Not all order arguments are accepted and this could result
        in an error from broker
        2) oms_order_id is mandatory for modifying orders
        """
        orders = self.pending_orders()
        orders = self.filter(orders, symbol=symbol, **kwargs)
        responses = []
        if len(orders) == 0:
            # Return in case of no matching orders
            return responses
        for order in orders:
            order_id = order["oms_order_id"]
            symbol = order["symbol"]
            resp = self.order_modify(oms_order_id=order_id, symbol=symbol, **kwargs)
            responses.append(resp)
        return responses

    def modify_bracket_stop(self, symbol, stop, first=False, p=0, n=None):
        """
        Modify stop loss value for bracket order
        symbol
            symbol to modify
        stop
            stop loss price to modify - actual stop loss
        first
            whether to modify the first order or all orders
            By default, all orders are modified
            If first=True, only the first order is modified
        p
            percentage of total open quantity to modify stop price
        n
            number of orders to be closed
        Note
        ----
        1) This implementation is exclusive to this broker - master trust
        2) stop is the actual stop loss price
        3) stop, target is identified by status
        4) if both p and n, percentage takes precedence
        """
        orders = self.pending_orders()
        orders = self.filter(
            orders, symbol=symbol, product="BO", status="trigger pending"
        )
        responses = []
        url = f"{self.base_url}/api/v1/orders"
        if len(orders) == 0:
            # Return in case of no matching orders
            return responses
        total_quantity = sum([o.get("quantity", 0) for o in orders])
        threshold_to_exit = total_quantity
        p = min(p, 100)
        if p > 0:
            threshold_to_exit = int(total_quantity * p * 0.01)
        qty = 0
        if n is None:
            n = len(orders)
        if p > 0:
            for order in orders:
                q = order.get("quantity", 0)
                qty += q
                kwargs = {
                    "oms_order_id": order["oms_order_id"],
                    "trading_symbol": order["symbol"],
                    "order_type": order["order_type"],
                    "exchange": order["exchange"],
                    "quantity": order["quantity"],
                    "product": order["product"],
                    "validity": order["validity"],
                    "instrument_token": order["instrument_token"],
                    "trigger_price": stop,
                    "price": stop,
                    "client_id": self.client_id,
                }
                payload = kwargs.copy()
                resp = requests.put(url, headers=self.headers, params=payload)
                responses.append(self._response(resp))
                if qty > threshold_to_exit:
                    return responses
            return responses

        # This code is run only when ou need to modify by number of orders
        for i, order in enumerate(orders):
            if i >= n:
                # Since the number of orders to be squared off is met,
                # we exit the program
                return responses
            kwargs = {
                "oms_order_id": order["oms_order_id"],
                "trading_symbol": order["symbol"],
                "order_type": order["order_type"],
                "exchange": order["exchange"],
                "quantity": order["quantity"],
                "product": order["product"],
                "validity": order["validity"],
                "instrument_token": order["instrument_token"],
                "trigger_price": stop,
                "price": stop,
                "client_id": self.client_id,
            }
            payload = kwargs.copy()
            resp = requests.put(url, headers=self.headers, params=payload)
            responses.append(self._response(resp))
            if first:
                return responses
        return responses

    def modify_bracket_target(self, symbol, target, first=False, n=None, p=0):
        """
        Modify target value for bracket order
        symbol
            symbol to modify
        stop
            target price to modify - actual stop loss
        first
            If True, close only the first order
        n
            number of orders to close
        Note
        ----
        1) This implementation is exclusive to this broker - master trust
        2) target is the actual target price
        3) stop, target is identified by status
        first
            whether to modify the first order or all orders
            By default, all orders are modified
            If first=True, only the first order is modified
        """
        orders = self.pending_orders()
        orders = self.filter(orders, symbol=symbol, product="BO", status="open")
        responses = []
        url = f"{self.base_url}/api/v1/orders"
        if len(orders) == 0:
            # Return in case of no matching orders
            return responses
        total_quantity = sum([o.get("quantity", 0) for o in orders])
        threshold_to_exit = total_quantity
        p = min(p, 100)
        if p > 0:
            threshold_to_exit = int(total_quantity * p * 0.01)
        qty = 0
        if n is None:
            n = len(orders)

        if p > 0:
            for order in orders:
                q = order.get("quantity", 0)
                qty += q
                kwargs = {
                    "oms_order_id": order["oms_order_id"],
                    "trading_symbol": order["symbol"],
                    "order_type": order["order_type"],
                    "exchange": order["exchange"],
                    "quantity": order["quantity"],
                    "product": order["product"],
                    "validity": order["validity"],
                    "instrument_token": order["instrument_token"],
                    "price": target,
                    "client_id": self.client_id,
                }
                payload = kwargs.copy()
                resp = requests.put(url, headers=self.headers, params=payload)
                responses.append(self._response(resp))
                if qty > threshold_to_exit:
                    return responses
            return responses
        for i, order in enumerate(orders):
            if i >= n:
                # Since the number of orders to be squared off is met,
                # we exit the program
                return responses
            kwargs = {
                "oms_order_id": order["oms_order_id"],
                "trading_symbol": order["symbol"],
                "order_type": order["order_type"],
                "exchange": order["exchange"],
                "quantity": order["quantity"],
                "product": order["product"],
                "validity": order["validity"],
                "instrument_token": order["instrument_token"],
                "price": target,
                "client_id": self.client_id,
            }
            payload = kwargs.copy()
            resp = requests.put(url, headers=self.headers, params=payload)
            responses.append(self._response(resp))
            if first:
                return responses
        return responses

    def exit_bracket_by_symbol(self, symbol, first=False, p=0):
        """
        Exit bracket order by symbol
        symbol
            symbol to exit bracket order
        first
            whether to modify the first order or all orders
            By default, all orders are modified
            If first=True, only the first order is modified
        """
        orders = self.pending_orders()
        orders = self.filter(orders, symbol=symbol, product="BO", status="open")
        responses = []
        if len(orders) == 0:
            # Return in case of no matching orders
            return responses
        total_quantity = sum([o.get("quantity", 0) for o in orders])
        threshold_to_exit = total_quantity
        p = min(p, 100)
        if p > 0:
            threshold_to_exit = int(total_quantity * p * 0.01)
            print(f"Exit threshold = {total_quantity}, {p}, {threshold_to_exit}")
        qty = 0
        for order in orders:
            oms_order_id = order["oms_order_id"]
            q = order.get("quantity", 0)
            qty += q
            leg_order_indicator = order["leg_order_indicator"]
            kwargs = {
                "oms_order_id": oms_order_id,
                "leg_order_indicator": leg_order_indicator,
                "status": "open",
                "client_id": self.client_id,
            }
            if leg_order_indicator:
                response = self.exit_bracket_order(**kwargs)
            else:
                response = self.order_cancel(oms_order_id)
            responses.append(response)
            if qty > threshold_to_exit:
                return responses
            if first:
                return responses
        return responses

    def modify_all_orders_by_conditions(
        self, modifications: Dict = None, n: int = 0, **kwargs
    ) -> List:
        """
        Modify all orders by the given condition
        """
        responses = []
        url = f"{self.base_url}/api/v1/orders"
        if not (modifications):
            return responses
        orders = self.pending_orders()
        orders = self.filter(orders, **kwargs)
        if len(orders) == 0:
            return responses
        if n <= 0:
            n = len(orders)
        for i, order in enumerate(orders):
            if i >= n:
                # Since the number of orders to be squared off is met,
                # we exit the program
                return responses
            kwargs = {
                "oms_order_id": order["oms_order_id"],
                "instrument_token": order["instrument_token"],
                "exchange": order["exchange"],
                "product": order["product"],
                "validity": order["validity"],
                "order_type": order["order_type"],
                "quantity": order["quantity"],
                "client_id": self.client_id,
            }
            kwargs.update(modifications)
            payload = kwargs.copy()
            resp = requests.put(url, headers=self.headers, params=payload)
            responses.append(self._response(resp))
        return responses

    def cancel_all_orders_by_conditions(self, n: int = 0, **kwargs) -> List:
        """
        Modify all orders by the given condition
        """
        responses = []
        url = f"{self.base_url}/api/v1/orders"
        orders = self.pending_orders()
        orders = self.filter(orders, **kwargs)
        if len(orders) == 0:
            return responses
        if n <= 0:
            n = len(orders)
        for i, order in enumerate(orders):
            if i >= n:
                # Since the number of orders to be squared off is met,
                # we exit the program
                return responses
            oms_order_id = order["oms_order_id"]
            resp = self.order_cancel(oms_order_id)
            responses.append(resp)
        return responses



from omspy.base import Broker, pre, post
from typing import Optional, List, Dict, Union
from ks_api_client import ks_api
import pendulum
import pandas as pd
import logging
from copy import deepcopy


def get_url(segment: Optional[str] = "cash") -> str:
    dt = pendulum.now(tz="Asia/Kolkata")
    date_string = dt.strftime("%d_%m_%Y")
    dct = {"cash": "Cash", "fno": "FNO"}
    seg = dct.get(segment, "Cash")
    url = f"https://preferred.kotaksecurities.com/security/production/TradeApiInstruments_{seg}_{date_string}.txt"
    return url


def get_name_for_cash_symbol(
    instrument_name: str, instrument_type: Optional[str] = None
):
    """
    Get the broker interface name for a symbol in cash segment
    """
    if instrument_type is None:
        pass
    elif pd.isna(instrument_type):
        instrument_type = None
    elif not (instrument_type.isalnum()):
        instrument_type = None
    elif instrument_type.upper() == "EQ":
        instrument_type = None
    elif instrument_type.lower() in ("na", "nan"):
        instrument_type = None

    if instrument_type is None:
        return instrument_name
    else:
        return f"{instrument_name}-{instrument_type or False}"


def get_name_for_fno_symbol(
    instrument_name: str,
    expiry: Union[pendulum.Date, pendulum.DateTime, str],
    option_type: Optional[str] = None,
    strike: Optional[Union[float, int]] = None,
):
    """
    Get the broker interface name for a future or option symbol
    Note
    -----
    1)If either the option type or option strike argument is not given, it is considered a futures contract
    2) A contract is considered an option contract only if both option type and strike price is provided
    """
    if isinstance(expiry, str):
        expiry = pendulum.parse(expiry)
    expiry = expiry.strftime("%d%b%y").upper()

    # Set the option type
    opts = {"ce": "call", "pe": "put"}
    if option_type is None:
        option_type = None
    elif pd.isna(option_type):
        option_type = None
    elif not (isinstance(option_type, str)):
        option_type = None
    elif option_type.lower() not in ("ce", "pe"):
        option_type = None
    else:
        option_type = opts.get(option_type.lower())

    # Set the strike
    if strike is None:
        strike = None
    elif not (isinstance(strike, (int, float))):
        strike = None
    elif strike <= 0:
        strike = None

    # Consider an instrument futures if there is no
    # corresponding strike or option type
    if (option_type is None) or (strike is None):
        return f"{instrument_name}{expiry}FUT".upper()
    else:
        return f"{instrument_name}{expiry}{strike}{option_type}".upper()


def convert_strike(strike: Union[int, float]) -> Union[int, float]:
    """
    Convert the type of the strike to an integer if it is an integer or float to 2 points if it is a float
    Note
    ----
    This is only a type conversion
    """
    if (strike - int(strike)) == 0:
        return int(strike)
    else:
        return round(strike, 2)


def download_file(url: str) -> pd.DataFrame:
    """
    Given a url, download the file, parse contents
    and return a dataframe.
    returns an empty Dataframe in case of an error
    """
    try:
        df = pd.read_csv(url, delimiter="|", parse_dates=["expiry"])
        df = df.rename(columns=lambda x: x.lower())
        return df.drop_duplicates(subset=["instrumenttoken"])
    except Exception as e:
        logging.error(e)
        return pd.DataFrame()


def add_name(data, segment: Optional[str] = "cash") -> pd.DataFrame:
    """
    add name to the given dataframe and return it
    data
        dataframe with the instrument data
    segment
        segment to add name to either cash or fno
    Note
    -----
    1) The extra column added is inst_name
    """
    if segment.lower() == "cash":
        data["inst_name"] = [
            f"{k}:{get_name_for_cash_symbol(x,y)}"
            for x, y, k in zip(
                data.instrumentname.values,
                data.instrumenttype.values,
                data.exchange.values,
            )
        ]
        return data
    elif segment.lower() == "fno":
        data["inst_name"] = [
            f"{k}:{get_name_for_fno_symbol(a,str(b),c,convert_strike(d))}"
            for a, b, c, d, k in zip(
                data.instrumentname.values,
                data.expiry.values,
                data.optiontype.values,
                data.strike.values,
                data.exchange.values,
            )
        ]
        return data
    else:
        return data


def create_instrument_master() -> Dict[str, int]:
    """
    Create the instrument master
    Note
    ----
    Takes no arguments and returns the entire instrument_master as a dictionary with key as name and values as instrument token
    """
    # TODO: Handle error in case of no instruments
    cash = download_file(get_url(segment="cash"))
    fno = download_file(get_url(segment="fno"))
    cash = add_name(cash, segment="cash")
    fno = add_name(fno, segment="fno")
    df = pd.concat([cash, fno]).drop_duplicates(subset=["instrumenttoken"])
    return {k: int(v) for k, v in zip(df.inst_name.values, df.instrumenttoken.values)}


class Kotak(Broker):
    """
    Automated Trading class
    """

    def __init__(
        self,
        access_token: str,
        userid: str,
        password: str,
        consumer_key: str,
        access_code: Optional[str] = None,
        instrument_master: Optional[str] = None,
        ip: str = "127.0.0.1",
        app_id: str = "default",
    ):
        self._access_token = access_token
        self._userid = userid
        self._password = password
        self._consumer_key = consumer_key
        self._access_code = access_code
        self._ip = ip
        self._app_id = app_id
        if instrument_master is None:
            self.master = create_instrument_master()
        else:
            self.master = instrument_master
        self._rev_master = {v: k for k, v in self.master.items()}
        super(Kotak, self).__init__()

    def get_instrument_token(self, instrument: str) -> Union[int, None]:
        return self.master.get(instrument)

    def _response(self, response: Dict) -> Union[Dict, None]:
        """
        return the response if success key is found else
        return None
        """
        if "Success" in response:
            return response["Success"]
        elif "success" in response:
            return response["success"]
        else:
            return None

    def _get_order_id(self, response: Optional[Dict] = None) -> Union[int, None]:
        """
        return order id from response
        """
        if response is None:
            return None
        elif isinstance(response, dict):
            response = self._response(response)
            if response is None:
                return None
            else:
                return dict(*response.values())["orderId"]
        else:
            return None

    def authenticate(self) -> None:
        try:
            client = ks_api.KSTradeApi(
                access_token=self._access_token,
                userid=self._userid,
                consumer_key=self._consumer_key,
                ip=self._ip,
                app_id=self._app_id,
            )
            client.login(password=self._password)
            client.session_2fa(access_code=self._access_code)
            self.client = client
        except Exception as e:
            logging.error(e)
            self.client = None

    def get_status(self, status: str) -> str:
        """
        Get the order status
        """
        status_map = dict(
            TRAD="COMPLETE",
            CAN="CANCELED",
            OPN="PENDING",
            SLO="PENDING",
            NEWF="PENDING",
        )
        return status_map.get(status.upper(), "PENDING")

    @property
    @post
    def orders(self) -> List[Dict]:
        """
        Return the orders for the day
        """
        order_report = self.client.order_report()
        if "success" in order_report:
            for order in order_report["success"]:
                order["symbol"] = self._rev_master.get(
                    order["instrumentToken"], order["instrumentName"]
                )
                order["status"] = self.get_status(order["status"])
            return order_report["success"]
        else:
            return order_report

    @property
    @post
    def positions(self) -> List[Dict]:
        """
        Return the positions only for the day
        """
        positions_report = self.client.positions(position_type="TODAYS")
        if "Success" in positions_report:
            for position in positions_report["Success"]:
                position["symbol"] = self._rev_master.get(
                    position["instrumentToken"], position["instrumentName"]
                )
            return positions_report["Success"]
        else:
            return positions_report

    def trades(self) -> List[Dict]:
        pass

    def _get_order_type(self) -> Dict:
        pass

    def order_place(
        self, symbol: str, side: str, exchange: str = "NSE", quantity: int = 1, **kwargs
    ) -> Union[str, None]:
        try:
            order_args = dict(
                variety="REGULAR", validity="GFD", order_type="MIS", price=0
            )
            token = self.get_instrument_token(f"{exchange}:{symbol}".upper())
            if not (token):
                logging.warning("No token for this symbol,check your symbol")
                return
            order_args.update(
                dict(
                    transaction_type=side.upper(),
                    instrument_token=token,
                    quantity=quantity,
                )
            )
            order_args.update(kwargs)
            response = self.client.place_order(**order_args)
            return self._get_order_id(response)
        except Exception as e:
            logging.error(e)
            return None

    def order_cancel(self, order_id: str) -> Dict:
        """
        Cancel an existing order
        """
        try:
            order_id = str(order_id)
            response = self.client.cancel_order(order_id=order_id)
            return self._get_order_id(response)
        except Exception as e:
            logging.error(e)
            return None

    def order_modify(self, order_id: str, **kwargs) -> Union[str, None]:
        """
        Modify an order
        """
        try:
            order_id = str(order_id)
            response = self.client.modify_order(order_id=order_id, **kwargs)
            response = self._response(response)
            return self._get_order_id(response)
        except Exception as e:
            logging.error(e)
            return None



"""
General utility functions for conversion and else
"""
from typing import Dict, Any, List
from omspy.models import BasicPosition
from collections import defaultdict


def create_basic_positions_from_orders_dict(
    orders: List[Dict],
) -> Dict[str, BasicPosition]:
    """
    Create a dictionary of positions from list of orders received from broker
    orders
        list of orders
    returns a dictionary with key being the symbol and values as a BasicPosition model
    """
    dct = {}
    for order in orders:
        symbol = order.get("symbol")
        if symbol:
            price = max(
                order.get("average_price", 0),
                order.get("price", 0),
                order.get("trigger_price", 0),
            )
            quantity = abs(order.get("quantity", 0))
            side = order.get("side").lower()
            if symbol not in dct.keys():
                # Create symbol if it doesn't exist
                dct[symbol] = BasicPosition(symbol=symbol)
            position = dct[symbol]
            if side == "buy":
                position.buy_quantity += quantity
                position.buy_value += price * quantity
            elif side == "sell":
                position.sell_quantity += quantity
                position.sell_value += price * quantity
    return dct


def dict_filter(lst: List[Dict], **kwargs) -> List[Dict]:
    """
    Filter a list of dictionary to conditions matching
    in kwargs
    lst
        list of dictionaries
    kwargs
        key values to filter; key is the dictionary key
        and value is the value to match.
        **This is an AND filter**

    Note
    -----
    For each dictionary in the list, each of the arguments
    in kwargs are matched and only those dictionaries that
    match all the conditions are returned
    """
    if len(lst) == 0:
        print("Nothing in the list")
        return []
    new_lst = []
    for d in lst:
        case = True
        for k, v in kwargs.items():
            if d.get(k) != v:
                case = False
        if case:
            new_lst.append(d)
    return new_lst


def tick(price, tick_size=0.05):
    """
    Rounds a given price to the requested tick
    """
    return round(price / tick_size) * tick_size


def stop_loss_step_decimal(
    price: float, side: str = "B", dec: float = 0.45, step: int = 2
) -> float:
    """
    Truncates down the stop loss value to the desired step
    and adds the given decimal
    price
        stop loss price
    side
        side to place order, the actual stop loss side
        B for BUY, S for SELL
    dec
        fixed decimal to be added
    step
        step size to determine stop
    Note
    ----
    1. Step object is always a positive number
    2. Side is the actual stop loss side you are placing the order
    """
    step = abs(step)
    m = int(price / step)
    val = (m + 1) * step if side == "S" else (m * step) - 1
    val = val + 1 - dec if side == "S" else val + dec
    return val



from omspy.brokers.api_helper import ShoonyaApiPy
from omspy.base import Broker, pre, post
from typing import Optional, List, Dict, Union
import pendulum


class Finvasia(Broker):
    """
    Automated Trading class
    """

    def __init__(
        self,
        user_id: str,
        password: str,
        pin: str,
        vendor_code: str,
        app_key: str,
        imei: str,
    ):
        self._user_id = user_id
        self._password = password
        self._pin = pin
        self._vendor_code = vendor_code
        self._app_key = app_key
        self._imei = imei
        self.finvasia = ShoonyaApiPy()
        super(Finvasia, self).__init__()

    def login(self) -> Union[Dict, None]:
        return self.finvasia.login(
            userid=self._user_id,
            password=self._password,
            twoFA=self._pin,
            vendor_code=self._vendor_code,
            api_secret=self._app_key,
            imei=self._imei,
        )

    def authenticate(self) -> Union[Dict, None]:
        """
        Authenticate the user
        """
        return self.login()

    def _convert_symbol(self, symbol: str, exchange: str = "NSE") -> str:
        """
        Convert raw symbol to finvasia
        """
        if symbol.endswith("-EQ") or symbol.endswith("-eq"):
            return symbol
        else:
            return f"{symbol}-EQ"

    @property
    @post
    def orders(self) -> List[Dict]:
        orderbook = self.finvasia.get_order_book()
        return orderbook

    @property
    @post
    def positions(self) -> List[Dict]:
        positionbook = self.finvasia.get_positions()
        return positionbook

    @property
    @post
    def trades(self) -> List[Dict]:
        tradebook = self.finvasia.get_trade_book()
        return tradebook

    def get_order_type(self, order_type: str) -> str:
        """
        Convert a generic order type to this specific
        broker's order type string
        returns MKT if the order_type is not matching
        """
        order_types = dict(LIMIT="LMT", MARKET="MKT", SL="SL-LMT", SLM="SL-MKT")
        return order_types.get(order_type.upper(), "MKT")

    @pre
    def order_place(self, **kwargs) -> Union[str, None]:
        symbol = kwargs.pop("symbol")
        symbol = self._convert_symbol(symbol)
        side = kwargs.pop("side")
        order_type = kwargs.pop("order_type", "MKT")
        if order_type:
            order_type = self.get_order_type(order_type)
        if side:
            side = side.upper()[0]
        if symbol:
            symbol = symbol.upper()
        order_args = dict(
            tradingsymbol=symbol,
            buy_or_sell=side,
            price_type=order_type,
            exchange="NSE",
            retention="DAY",
            product_type="I",
            discloseqty=0,
        )
        order_args.update(kwargs)
        return self.finvasia.place_order(**order_args)

    def order_cancel(self, order_id: str) -> Union[Dict, None]:
        """
        Cancel an existing order
        """
        return self.finvasia.cancel_order(orderno=order_id)

    @pre
    def order_modify(self, **kwargs) -> Union[str, None]:
        """
        Modify an existing order
        """
        symbol = kwargs.pop("tradingsymbol")
        order_id = kwargs.pop("order_id", None)
        order_type = kwargs.pop("order_type", "MKT")
        if order_type:
            order_type = self.get_order_type(order_type)
        if symbol:
            symbol = self._convert_symbol(symbol).upper()
        order_args = dict(
            orderno=order_id,
            newprice_type=order_type,
            exchange="NSE",
            tradingsymbol=symbol,
        )
        order_args.update(kwargs)
        return self.finvasia.modify_order(**order_args)



from NorenRestApiPy.NorenApi import  NorenApi
from threading import Timer
import pandas as pd
import time
import concurrent.futures

api = None
class Order:
     def __init__(self, buy_or_sell:str = None, product_type:str = None,
                 exchange: str = None, tradingsymbol:str =None, 
                 price_type: str = None, quantity: int = None, 
                 price: float = None,trigger_price:float = None, discloseqty: int = 0,
                 retention:str = 'DAY', remarks: str = "tag",
                 order_id:str = None):
        self.buy_or_sell=buy_or_sell
        self.product_type=product_type
        self.exchange=exchange
        self.tradingsymbol=tradingsymbol
        self.quantity=quantity
        self.discloseqty=discloseqty
        self.price_type=price_type
        self.price=price
        self.trigger_price=trigger_price
        self.retention=retention
        self.remarks=remarks
        self.order_id=None


    #print(ret)

    


def get_time(time_string):
    data = time.strptime(time_string,'%d-%m-%Y %H:%M:%S')

    return time.mktime(data)


class ShoonyaApiPy(NorenApi):
    def __init__(self):
        NorenApi.__init__(self, host='https://shoonyatrade.finvasia.com/NorenWClientTP/', websocket='wss://shoonyatrade.finvasia.com/NorenWSTP/', eodhost='https://shoonya.finvasia.com/chartApi/getdata/')
        global api
        api = self

    def place_basket(self, orders):

        resp_err = 0
        resp_ok  = 0
        result   = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:

            future_to_url = {executor.submit(self.place_order, order): order for order in  orders}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
            try:
                result.append(future.result())
            except Exception as exc:
                print(exc)
                resp_err = resp_err + 1
            else:
                resp_ok = resp_ok + 1

        return result
                
    def placeOrder(self,order: Order):
        ret = NorenApi.place_order(self, buy_or_sell=order.buy_or_sell, product_type=order.product_type,
                            exchange=order.exchange, tradingsymbol=order.tradingsymbol, 
                            quantity=order.quantity, discloseqty=order.discloseqty, price_type=order.price_type, 
                            price=order.price, trigger_price=order.trigger_price,
                            retention=order.retention, remarks=order.remarks)
        #print(ret)

        return ret



"""
This module contains the list of basic models
"""
from pydantic import BaseModel, validator, PrivateAttr
from typing import Optional, List
import pendulum


class QuantityMatch(BaseModel):
    buy: int = 0
    sell: int = 0

    @property
    def is_equal(self) -> bool:
        return True if self.buy == self.sell else False

    @property
    def not_matched(self) -> int:
        return self.buy - self.sell


class BasicPosition(BaseModel):
    """
    A simple position tracking mechanism
    """

    symbol: str
    buy_quantity: int = 0
    sell_quantity: int = 0
    buy_value: float = 0.0
    sell_value: float = 0.0

    @property
    def net_quantity(self) -> int:
        return self.buy_quantity - self.sell_quantity

    @property
    def average_buy_value(self) -> float:
        return self.buy_value / self.buy_quantity if self.buy_value > 0.0 else 0.0

    @property
    def average_sell_value(self) -> float:
        return self.sell_value / self.sell_quantity if self.sell_quantity > 0 else 0.0


class Quote(BaseModel):
    price: float
    quantity: int
    orders: Optional[int] = None

    @property
    def value(self):
        return self.price * self.quantity


class OrderBook(BaseModel):
    bid: List[Quote]
    ask: List[Quote]


class Tracker(BaseModel):
    """
    A simple tracker to track the high low prices
    """

    name: str  # name of the symbol to be tracked
    last_price: float = 0
    # Initializing high and low to extreme values
    high: float = -1e100
    low: float = 1e100

    def update(self, last_price: float):
        self.last_price = last_price
        self.high = max(last_price, self.high)
        self.low = min(last_price, self.low)


class Timer(BaseModel):
    """
    A simple timer that could be attached to any model
    """

    start_time: pendulum.DateTime
    end_time: pendulum.DateTime
    timezone: Optional[str] = None

    @validator("end_time")
    def validate_times(cls, v, values):
        """
        Validate end_time greater than start_time
        and start_time greater than current time
        """
        start = values.get("start_time")
        tz = values.get("timezone")
        if v < start:
            raise ValueError("end time greater than start time")
        if start < pendulum.now(tz=tz):
            raise ValueError("start time lesser than current time")
        return v

    @property
    def has_started(self):
        """
        Whether tracking has started
        """
        return True if pendulum.now(tz=self.timezone) > self.start_time else False

    @property
    def has_completed(self):
        """
        Whether tracking has completed
        """
        return True if pendulum.now(tz=self.timezone) > self.end_time else False


class TimeTracker(Tracker, Timer):
    pass


class OrderLock(BaseModel):
    """
    Lock order placement, modification and cancellation
    for a few seconds
    """

    max_order_creation_lock_time: float = 60
    max_order_modification_lock_time: float = 60
    max_order_cancellation_lock_time: float = 60
    timezone: Optional[str] = None
    _creation_lock_till: pendulum.DateTime = PrivateAttr()
    _modification_lock_till: pendulum.DateTime = PrivateAttr()
    _cancellation_lock_till: pendulum.DateTime = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._creation_lock_till = pendulum.now(tz=self.timezone)
        self._modification_lock_till = pendulum.now(tz=self.timezone)
        self._cancellation_lock_till = pendulum.now(tz=self.timezone)

    @property
    def creation_lock_till(self):
        return self._creation_lock_till

    @property
    def modification_lock_till(self):
        return self._modification_lock_till

    @property
    def cancellation_lock_till(self):
        return self._cancellation_lock_till

    def create(self, seconds: float) -> pendulum.DateTime:
        """
        Lock the create_order function for the given number of seconds
        """
        seconds = min(seconds, self.max_order_creation_lock_time)
        self._creation_lock_till = pendulum.now(tz=self.timezone).add(seconds=seconds)
        return self.creation_lock_till

    def modify(self, seconds: float) -> pendulum.DateTime:
        """
        Lock the modify_order function for the given number of seconds
        """
        seconds = min(seconds, self.max_order_modification_lock_time)
        self._modification_lock_till = pendulum.now(tz=self.timezone).add(
            seconds=seconds
        )
        return self.modification_lock_till

    def cancel(self, seconds: float) -> pendulum.DateTime:
        """
        Lock the cancel_order function for the given number of seconds
        """
        seconds = min(seconds, self.max_order_cancellation_lock_time)
        self._cancellation_lock_till = pendulum.now(tz=self.timezone).add(
            seconds=seconds
        )
        return self.cancellation_lock_till

    @property
    def can_create(self) -> bool:
        """
        returns True if order can be created else False
        """
        return (
            True if pendulum.now(tz=self.timezone) > self.creation_lock_till else False
        )

    @property
    def can_modify(self) -> bool:
        """
        returns True if order can be modified else False
        """
        return (
            True
            if pendulum.now(tz=self.timezone) > self.modification_lock_till
            else False
        )

    @property
    def can_cancel(self) -> bool:
        """
        returns True is order can be canceled else False
        """
        return (
            True
            if pendulum.now(tz=self.timezone) > self.cancellation_lock_till
            else False
        )



from pydantic import BaseModel, ValidationError, validator
from typing import List, Optional, Type, Tuple, Union
from omspy.base import Broker
from omspy.order import Order, CompoundOrder
from omspy.models import Timer
import pendulum
import sqlite_utils
import logging


class BaseStrategy(BaseModel):
    start_time: pendulum.DateTime
    end_time: pendulum.DateTime
    cycle: int = 0
    done: bool = False
    broker: Optional[Type[Broker]] = None
    connection: Optional[sqlite_utils.Database] = None
    timezone: Optional[str] = None
    _timer: Optional[Timer] = None

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True

    def __init__(self, **data):
        super().__init__(**data)
        self._timer = Timer(
            start_time=self.start_time, end_time=self.end_time, timezone=self.timezone
        )

    @property
    def timer(self):
        return self._timer


class ShortStraddle(BaseStrategy):
    symbols: Tuple[str, str]
    limit_price: Optional[Tuple[float, float]] = (0.0, 0.0)
    trigger_price: Optional[Tuple[float, float]] = (0.0, 0.0)
    stop_price: Optional[Tuple[float, float]] = (0.0, 0.0)
    quantity: int = 1
    exclude_stop: bool = False
    _order: Optional[CompoundOrder] = None

    def __init__(self, **data):
        super().__init__(**data)
        self._order = CompoundOrder(
            broker=self.broker, connection=self.connection, timezone=self.timezone
        )

    @property
    def order(self) -> CompoundOrder:
        return self._order

    def create_order(self):
        com = self._order
        s1, s2 = self.symbols
        order1 = Order(symbol=s1, side="sell", quantity=self.quantity)
        order1.price = self.limit_price[0]
        order2 = order1.clone()
        order2.symbol = s2
        order2.price = self.limit_price[1]
        com.add(order1)
        com.add(order2)
        order1stop = order1.clone()
        order1stop.trigger_price = self.trigger_price[0]
        order1stop.price = self.stop_price[0]
        order1stop.order_type = "SL"
        order1stop.side = "buy"
        com.add(order1stop)
        order2stop = order2.clone()
        order2stop.trigger_price = self.trigger_price[1]
        order2stop.price = self.stop_price[1]
        order2stop.order_type = "SL"
        order2stop.side = "buy"
        com.add(order2stop)
        return self.order

    def check_sell_both_sides(self) -> bool:
        """
        Check whether the sell order is placed on both the
        sides and the orders are completed
        """
        orders = self.order.orders
        if orders[0].is_complete and orders[1].is_complete:
            return True



from typing import Optional, Dict, List, Type, Any, Union, Tuple, Callable
from omspy.base import Broker
from omspy.order import Order, CompoundOrder
from omspy.models import OrderLock
import pendulum
from pydantic import BaseModel, ValidationError, validator
import logging


class BasicPeg(CompoundOrder):
    symbol: str
    side: str
    quantity: int
    timezone: str = "UTC"

    def __init__(self, **data) -> None:
        super().__init__(**data)
        pop_attribs = [
            "symbol",
            "side",
            "quantity",
            "timezone",
            "order_type",
            "connection",
        ]
        for attrib in pop_attribs:
            if attrib in data:
                data.pop(attrib)
        order = Order(
            symbol=self.symbol,
            quantity=self.quantity,
            side=self.side,
            timezone=self.timezone,
            order_type="LIMIT",
            **data
        )
        self.add(order)
        self.ltp[self.symbol] = 0


class PegMarket(BasicPeg):
    duration: int = 60
    peg_every: int = 10
    convert_to_market_after_expiry = True
    _next_peg: Optional[pendulum.DateTime]
    _num_pegs: int = 0
    _max_pegs: int = 0
    _expire_at: Optional[pendulum.DateTime]

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self._max_pegs = int(self.duration / self.peg_every)
        self._num_pegs = 0
        self._expire_at = pendulum.now(tz=self.timezone).add(seconds=self.duration)
        self._next_peg = pendulum.now(tz=self.timezone).add(seconds=self.peg_every)

    def execute(self):
        self.orders[0].price = self.ref_price
        self.orders[0].execute(broker=self.broker, **self.order_args)

    @property
    def next_peg(self):
        return self._next_peg

    @property
    def num_pegs(self):
        return self._num_pegs

    @property
    def ref_price(self):
        return self.ltp.get(self.symbol)

    def run(self):
        order = self.orders[0]
        now = pendulum.now(self.timezone)
        if order.is_pending:
            if now > self.next_peg:
                self._next_peg = now.add(seconds=self.peg_every)
                order.modify(broker=self.broker, price=self.ref_price)
            if now > self._expire_at:
                if self.convert_to_market_after_expiry:
                    order.modify(broker=self.broker, order_type="MARKET")
                else:
                    order.cancel(self.broker)


class PegExisting(BaseModel):
    order: Order
    broker: Any
    timezone: Optional[str] = None
    duration: int = 60
    peg_every: int = 10
    done: bool = False
    lock: Optional[OrderLock] = None
    lock_duration: int = 2
    _next_peg: Optional[pendulum.DateTime] = None
    _num_pegs: int = 0
    _max_pegs: int = 0
    _expire_at: Optional[pendulum.DateTime]

    class Config:
        underscore_attrs_are_private = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self._max_pegs = int(self.duration / self.peg_every)
        self._num_pegs = 0
        self._expire_at = pendulum.now(tz=self.timezone).add(seconds=self.duration)
        self._next_peg = pendulum.now(tz=self.timezone).add(seconds=self.peg_every)
        self.order.order_type = "LIMIT"
        if self.lock is None:
            self.lock = OrderLock()

    @validator("order")
    def order_should_be_pending(cls, v):
        """
        Only accept a pending order
        """
        if not (v.is_pending):
            raise ValueError
        else:
            return v

    @property
    def next_peg(self) -> pendulum.DateTime:
        return self._next_peg

    @property
    def num_pegs(self) -> int:
        return self._num_pegs

    def execute(self, **order_args):
        self.order.execute(broker=self.broker)

    def _mark_done(self) -> bool:
        """
        Mark whether the order is done
        """
        if self.order.is_complete:
            self.done = True
        elif not (self.order.is_pending):
            self.done = True
        return self.done

    def run(self, ltp: float):
        if self.done:
            logging.warning("Order already done")
            return

        order = self.order
        now = pendulum.now(self.timezone)
        if order.is_pending:
            if now > self._expire_at:
                if self.order.convert_to_market_after_expiry:
                    if self.lock.can_modify:
                        order.modify(broker=self.broker, order_type="MARKET")
                        self.lock.modify(self.lock_duration)
                else:
                    if self.lock.can_cancel:
                        order.cancel(self.broker)
                        self.lock.cancel(self.lock_duration)
            elif now > self.next_peg:
                self._next_peg = now.add(seconds=self.peg_every)
                if self.lock.can_modify:
                    order.modify(broker=self.broker, price=ltp)
                    self.lock.modify(self.lock_duration)
        self._mark_done()


class PegSequential(BaseModel):
    """
    Peg orders in sequence and peg only when the
    previous order is complete
    """

    orders: List[Order]
    broker: Any
    timezone: Optional[str] = None
    duration: int = 12
    peg_every: int = 4
    lock_duration: int = 2
    done: bool = False
    _order: Optional[PegExisting] = None
    _start_time: Optional[pendulum.DateTime] = None

    class Config:
        underscore_attrs_are_private = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        # Validate whether orders could be pegged
        for order in self.orders:
            peg = PegExisting(
                order=order,
                timezone=self.timezone,
                duration=self.duration,
                peg_every=self.peg_every,
                lock_duration=self.lock_duration,
            )
        self._start_time = pendulum.now(tz=self.timezone)

    @property
    def has_expired(self) -> bool:
        """
        Whether the entire sequential order has expired.
        A sequential order is considered expired when the
        current time is greater than the maximum possible
        time taken by all the orders together
        Note
        ----
        1) total_time = duration*number of orders
        """
        total_time = self.duration * len(self.orders)
        if pendulum.now(tz=self.timezone) > self._start_time.add(seconds=total_time):
            return True
        else:
            return False

    @property
    def order(self) -> Optional[PegExisting]:
        return self._order

    @property
    def completed(self) -> List[Order]:
        """
        returns the list of completed orders
        """
        return [order for order in self.orders if order.is_complete]

    @property
    def pending(self):
        """
        returns the list of pending orders
        """
        return [order for order in self.orders if order.is_pending]

    @property
    def all_complete(self):
        """
        Whether all orders are completed
        """
        return all([order.is_complete for order in self.orders])

    def get_current_order(self) -> Union[PegExisting, None]:
        """
        Get the current order to peg
        """
        for order in self.orders:
            if order.is_pending:
                return PegExisting(
                    order=order,
                    broker=self.broker,
                    timezone=self.timezone,
                    duration=self.duration,
                    peg_every=self.peg_every,
                    lock_duration=self.lock_duration,
                )
        return None

    def set_current_order(self) -> Union[PegExisting, None]:
        """
        Set the current order for pegging
        Note
        ----
        1) Set the current order only when a existing peg
        order is complete or there is no peg order
        """
        if self.order is None:
            self._order = self.get_current_order()
        elif self.order.order.is_complete:
            self._order = self.get_current_order()
        return self.order

    def execute_all(self):
        # Execute all pending orders
        for order in self.orders:
            order.execute(broker=self.broker)

    def cancel_all(self):
        # Cancel all pending orders
        for order in self.orders:
            order.cancel(broker=self.broker)

    def run(self, ltp: Dict[str, float]):
        self.set_current_order()
        peg = self.order
        # Run only when there is an order
        if peg:
            if not (peg.order.order_id):
                # Place the order if it has not been placed yet
                peg.execute()
            symbol = peg.order.symbol
            last_price = ltp.get(symbol, 0)
            peg.run(ltp=last_price)

    def _mark_done(self):
        pass



from datetime import timezone
from typing import Optional, Dict, List, Type, Any, Union, Tuple, Callable
from omspy.base import Broker
from omspy.order import Order, CompoundOrder


class StopOrder(CompoundOrder):
    symbol: str
    side: str
    trigger_price: float
    price: float = 0.0
    quantity: int = 1
    disclosed_quantity: int = 0
    order_type: Tuple = ("LIMIT", "SL-M")

    def __init__(self, **data):
        super().__init__(**data)
        side_map = {"buy": "sell", "sell": "buy"}
        base_order = Order(
            symbol=self.symbol,
            side=self.side,
            quantity=self.quantity,
            disclosed_quantity=self.disclosed_quantity,
            order_type=self.order_type[0],
            price=self.price,
            trigger_price=self.trigger_price,
        )

        cover_order = base_order.clone()
        cover_order.order_type = self.order_type[1]
        cover_order.side = side_map.get(cover_order.side)
        self.add(base_order)
        self.add(cover_order)


class StopLimitOrder(CompoundOrder):
    def __init__(
        self,
        symbol: str,
        side: str,
        trigger_price: float,
        price: float = 0.0,
        stop_limit_price: float = 0.0,
        quantity: int = 1,
        order_type="MARKET",
        disclosed_quantity: int = 0,
        **kwargs,
    ):
        super(StopLimitOrder, self).__init__(**kwargs)
        side2 = "sell" if side.lower() == "buy" else "buy"
        if stop_limit_price == 0:
            stop_limit_price = trigger_price
        self.add_order(
            symbol=symbol,
            side=side,
            price=price,
            quantity=quantity,
            order_type=order_type,
            disclosed_quantity=disclosed_quantity,
        )
        self.add_order(
            symbol=symbol,
            side=side2,
            price=stop_limit_price,
            trigger_price=trigger_price,
            quantity=quantity,
            order_type="SL",
            disclosed_quantity=disclosed_quantity,
        )


class BracketOrder(StopOrder):
    def __init__(self, target: float, **kwargs):
        super(BracketOrder, self).__init__(**kwargs)
        self._target = target

    @property
    def target(self) -> float:
        return self._target

    @property
    def is_target_hit(self) -> bool:
        """
        Check whether the given target is hit
        """
        for k, v in self.ltp.items():
            # We assume a single symbol only so breaking
            # TO DO: A better way is appreciated
            ltp = v
            break
        return True if ltp > self.target else False

    def do_target(self) -> None:
        """
        Execute target order if target is hit
        Note
        -----
        This checks
         1. whether the target is hit
         2. if target is hit, modify the existing stop and exit the order
        """
        if self.is_target_hit:
            order = self.orders[-1]
            order.order_type = "MARKET"
            order.modify(broker=self.broker)


class TrailingStopOrder(StopLimitOrder):
    """
    Trailing stop order
    """

    def __init__(self, trail_by: Tuple[float, float], **kwargs):
        super(TrailingStopOrder, self).__init__(**kwargs)
        self.trail_big: float = trail_by[0]
        self.trail_small: float = trail_by[-1]
        self._maxmtm: float = 0
        self._stop: float = kwargs.get("trigger_price", 0)
        self.initial_stop = self._stop
        self.symbol: str = kwargs.get("symbol")
        self.quantity: int = kwargs.get("quantity", 1)

    @property
    def stop(self):
        return self._stop

    @property
    def maxmtm(self):
        return self._maxmtm

    def _update_maxmtm(self):
        self._maxmtm = max(self.total_mtm, self._maxmtm)

    def _update_stop(self):
        mtm_per_unit = self.maxmtm / self.quantity
        multiplier = self.trail_small / self.trail_big
        self._stop = self.initial_stop + (mtm_per_unit * multiplier)

    def watch(self):
        self._update_maxmtm()
        self._update_stop()
        ltp = self.ltp.get(self.symbol)
        if ltp:
            # TODO: Implement for sell also
            if ltp < self.stop:
                order = self.orders[-1]
                order.order_type = "MARKET"
                order.modify(broker=self.broker)



from pydantic import BaseModel, validator, Field, PrivateAttr, Json
from datetime import timezone
from typing import Optional, Dict, List, Type, Any, Union, Tuple, Callable, Set
import uuid
import pendulum
import sqlite3
import logging
from collections import Counter, defaultdict
from omspy.base import *
from copy import deepcopy
from sqlite_utils import Database


def get_option(spot: float, num: int = 0, step: float = 100.0) -> float:
    """
    Get the option price given number of strikes
    spot
        spot price of the instrument
    num
        number of strikes farther
    step
        step size of the option
    Note
    ----
    1. By default, the ATM option is fetched
    """
    v = round(spot / step)
    return v * (step + num)


def create_db(dbname: str = ":memory:") -> Union[Database, None]:
    """
    Create a sqlite3 database for the orders and return the connection
    dbname
        name of the database
        default in-memory database
    """
    try:
        con = sqlite3.connect(dbname)
        with con:
            con.execute(
                """create table orders
                           (
                           symbol text, side text, quantity integer,
                           id text primary key, parent_id text, timestamp text,
                           order_type text, broker_timestamp text,
                           exchange_timestamp text, order_id text,
                           exchange_order_id text, price real,
                           trigger_price real, average_price real,
                           pending_quantity integer, filled_quantity integer,
                           cancelled_quantity integer, disclosed_quantity integer,
                           validity text, status text,
                           expires_in integer, timezone text,
                           client_id text, convert_to_market_after_expiry text,
                           cancel_after_expiry text, retries integer, max_modifications integer,
                           exchange text, tag string, can_peg integer,
                           pseudo_id string, strategy_id string, portfolio_id string,
                           JSON text, error text, is_multi integer,
                           last_updated_at text
                           )"""
            )
            return Database(con)
    except Exception as e:
        logging.error(e)
        return None


class Order(BaseModel):
    symbol: str
    side: str
    quantity: int = 1
    id: Optional[str] = None
    parent_id: Optional[str] = None
    timestamp: Optional[pendulum.DateTime] = None
    order_type: str = "MARKET"
    broker_timestamp: Optional[pendulum.DateTime] = None
    exchange_timestamp: Optional[pendulum.DateTime] = None
    order_id: Optional[str] = None
    exchange_order_id: Optional[str] = None
    price: Optional[float] = None
    trigger_price: float = 0.0
    average_price: float = 0.0
    pending_quantity: Optional[int] = None
    filled_quantity: int = 0
    cancelled_quantity: int = 0
    disclosed_quantity: int = 0
    validity: str = "DAY"
    status: Optional[str] = None
    expires_in: int = 0
    timezone: str = "UTC"
    client_id: Optional[str] = None
    convert_to_market_after_expiry: bool = False
    cancel_after_expiry: bool = True
    retries: int = 0
    max_modifications: int = 10
    exchange: Optional[str] = None
    tag: Optional[str] = None
    connection: Optional[Database] = None
    can_peg: bool = True
    pseudo_id: Optional[str] = None
    strategy_id: Optional[str] = None
    portfolio_id: Optional[str] = None
    JSON: Optional[Json] = None
    error: Optional[str] = None
    is_multi: bool = False
    last_updated_at: Optional[pendulum.DateTime] = None
    _num_modifications: int = 0
    _attrs: Tuple[str] = (
        "exchange_timestamp",
        "exchange_order_id",
        "status",
        "filled_quantity",
        "pending_quantity",
        "disclosed_quantity",
        "average_price",
    )
    _exclude_fields: Set[str] = {"connection"}

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        from omspy.base import Broker

        if not (self.id):
            self.id = uuid.uuid4().hex
        tz = self.timezone
        if not (self.timestamp):
            self.timestamp = pendulum.now(tz=tz)
        self.pending_quantity = self.quantity
        if self.expires_in == 0:
            self.expires_in = (
                pendulum.today(tz=tz).end_of("day") - pendulum.now(tz=tz)
            ).seconds
        else:
            self.expires_in = abs(self.expires_in)

    @validator("quantity", always=True, allow_reuse=True)
    def quantity_not_negative(cls, v):
        if v < 0:
            raise ValueError("quantity must be positive")
        return v

    @property
    def is_complete(self) -> bool:
        if self.quantity == self.filled_quantity:
            return True
        elif self.status == "COMPLETE":
            return True
        elif (self.filled_quantity + self.cancelled_quantity) == self.quantity:
            return True
        else:
            return False

    @property
    def is_pending(self) -> bool:
        quantity = self.filled_quantity + self.cancelled_quantity
        # Order not pending if it is complete/canceled or rejected
        # irrespective of the filled and remaining quantity
        if self.status in ("COMPLETE", "CANCELED", "REJECTED"):
            return False
        elif quantity < self.quantity:
            return True
        else:
            return False

    @property
    def time_to_expiry(self) -> int:
        now = pendulum.now(tz=self.timezone)
        ts = self.timestamp
        return max(0, self.expires_in - (now - ts).seconds)

    @property
    def time_after_expiry(self) -> int:
        now = pendulum.now(tz=self.timezone)
        ts = self.timestamp
        return max(0, (now - ts).seconds - self.expires_in)

    @property
    def has_expired(self) -> bool:
        return True if self.time_to_expiry == 0 else False

    @property
    def has_parent(self) -> bool:
        return True if self.parent_id else False

    def update(self, data: Dict[str, Any], save: bool = True) -> bool:
        """
        Update order based on information received from broker
        data
            data to update as dictionary
        returns True if update is done
        Note
        ----
        1) Information is updated only for those keys specified in attrs
        2) Information is updated only when the order is not completed
        """
        if not (self.is_complete):
            for att in self._attrs:
                val = data.get(att)
                if val:
                    setattr(self, att, val)
            self.last_updated_at = pendulum.now(tz=self.timezone)
            if self.connection and save:
                self.save_to_db()
            return True
        else:
            return False

    def execute(self, broker: Any, **kwargs) -> Optional[str]:
        """
        Execute an order on a broker, place a new order
        kwargs
            Additional arguments to the order
        Note
        ----
        Only new arguments added to the order in keyword arguments
        """
        # Do not place a new order if this order is complete or has order_id
        from omspy.base import Broker as base_broker

        if not (self.is_complete) and not (self.order_id):
            order_args = {
                "symbol": self.symbol.upper(),
                "side": self.side.upper(),
                "order_type": self.order_type.upper(),
                "quantity": self.quantity,
                "price": self.price,
                "trigger_price": self.trigger_price,
                "disclosed_quantity": self.disclosed_quantity,
            }
            dct = {k: v for k, v in kwargs.items() if k not in order_args.keys()}
            order_args.update(dct)
            order_id = broker.order_place(**order_args)
            self.order_id = order_id
            if self.connection:
                self.save_to_db()
            return order_id
        else:
            return self.order_id

    def modify(self, broker: Any, **kwargs):
        """
        Modify an existing order
        """
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
        order_args = {
            "order_id": self.order_id,
            "quantity": self.quantity,
            "price": self.price,
            "trigger_price": self.trigger_price,
            "order_type": self.order_type.upper(),
            "disclosed_quantity": self.disclosed_quantity,
        }
        if self._num_modifications < self.max_modifications:
            broker.order_modify(**order_args)
            self._num_modifications += 1
        else:
            logging.info(f"Maximum number of modifications exceeded")

    def cancel(self, broker: Any):
        """
        Cancel an existing order
        """
        broker.order_cancel(order_id=self.order_id)

    def save_to_db(self) -> bool:
        """
        save or update the order to db
        """
        if self.connection:
            values = self.dict(exclude=self._exclude_fields)
            self.connection["orders"].upsert(values, pk="id")
            return True
        else:
            logging.info("No valid database connection")
            return False

    def clone(self):
        """
        Clone the order with a new order id
        Note
        ----
        returns a copy of the new order with a new
        order_id. parent_id is not copied
        """
        dct = self.dict(exclude={"id", "parent_id", "timestamp"})
        order = Order(**dct)
        return order


class CompoundOrder(BaseModel):
    broker: Any
    id: Optional[str] = None
    ltp: defaultdict = Field(default_factory=defaultdict)
    orders: List[Order] = Field(default_factory=list)
    connection: Optional[Database] = None
    order_args: Optional[Dict] = None

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        if not (self.id):
            self.id = uuid.uuid4().hex
        if self.order_args is None:
            self.order_args = {}

    @property
    def count(self) -> int:
        """
        return the number of orders
        """
        return len(self.orders)

    @property
    def positions(self) -> Counter:
        """
        return the positions as a dictionary
        """
        c: Counter = Counter()
        for order in self.orders:
            symbol = order.symbol
            qty = order.filled_quantity
            side = str(order.side).lower()
            sign = -1 if side == "sell" else 1
            qty = qty * sign
            c.update({symbol: qty})
        return c

    def add_order(self, **kwargs) -> Optional[str]:
        kwargs["parent_id"] = self.id
        if not (kwargs.get("connection")):
            kwargs["connection"] = self.connection
        order = Order(**kwargs)
        order.save_to_db()
        self.orders.append(order)
        return order.id

    def _average_price(self, side: str = "buy") -> Dict[str, float]:
        """
        Get the average price for all the instruments
        side
            side to calculate average price - buy or sel
        """
        side = str(side).lower()
        value_counter: Counter = Counter()
        quantity_counter: Counter = Counter()
        for order in self.orders:
            order_side = str(order.side).lower()
            if side == order_side:
                symbol = order.symbol
                price = order.average_price
                quantity = order.filled_quantity
                value = price * quantity
                value_counter.update({symbol: value})
                quantity_counter.update({symbol: quantity})
        dct: defaultdict = defaultdict()
        for v in value_counter:
            numerator = value_counter.get(v)
            denominator = quantity_counter.get(v)
            if numerator and denominator:
                dct[v] = numerator / denominator
        return dct

    @property
    def average_buy_price(self) -> Dict[str, float]:
        return self._average_price(side="buy")

    @property
    def average_sell_price(self) -> Dict[str, float]:
        return self._average_price(side="sell")

    def update_orders(self, data: Dict[str, Dict[str, Any]]) -> Dict[str, bool]:
        """
        Update all orders
        data
            data as dictionary with key as broker order_id
        returns a dictionary with order_id and update status as boolean
        """
        dct: Dict[str, bool] = {}
        for order in self.pending_orders:
            order_id = str(order.order_id)
            status = order.status
            if order_id in data:
                d = data.get(order_id)
                if d:
                    order.update(d)
                    dct[order_id] = True
                else:
                    dct[order_id] = False
            else:
                dct[order_id] = False
        return dct

    def _total_quantity(self) -> Dict[str, Counter]:
        """
        Get the total buy and sell quantity by symbol
        """
        buy_counter: Counter = Counter()
        sell_counter: Counter = Counter()
        for order in self.orders:
            side = order.side.lower()
            symbol = order.symbol
            quantity = abs(order.filled_quantity)
            if side == "buy":
                buy_counter.update({symbol: quantity})
            elif side == "sell":
                sell_counter.update({symbol: quantity})
        return {"buy": buy_counter, "sell": sell_counter}

    @property
    def buy_quantity(self) -> Counter:
        return self._total_quantity()["buy"]

    @property
    def sell_quantity(self) -> Counter:
        return self._total_quantity()["sell"]

    def update_ltp(self, last_price: Dict[str, float]):
        """
        Update ltp for the given symbols
        last_price
            dictionary with symbol as key and last price as value
        returns the ltp for all the symbols
        Note
        ----
        1. Last price is updated for all given symbols irrespective of
        orders placed
        """
        for symbol, ltp in last_price.items():
            self.ltp[symbol] = ltp
        return self.ltp

    @property
    def net_value(self) -> Counter:
        """
        Return the net value by symbol
        """
        c: Counter = Counter()
        for order in self.orders:
            symbol = order.symbol
            side = str(order.side).lower()
            sign = -1 if side == "sell" else 1
            value = order.filled_quantity * order.average_price * sign
            c.update({symbol: value})
        return c

    @property
    def mtm(self) -> Counter:
        c: Counter = Counter()
        net_value = self.net_value
        positions = self.positions
        ltp = self.ltp
        for symbol, value in net_value.items():
            c.update({symbol: -value})
        for symbol, quantity in positions.items():
            v = quantity * ltp.get(symbol, 0)
            c.update({symbol: v})
        return c

    @property
    def total_mtm(self) -> float:
        return sum(self.mtm.values())

    def execute_all(self, **kwargs):
        for order in self.orders:
            order_args = deepcopy(self.order_args)
            order_args.update(kwargs)
            order.execute(broker=self.broker, **order_args)

    def check_flags(self):
        """
        Check for flags on each order and take suitable action
        """
        for order in self.orders:
            if (order.is_pending) and (order.has_expired):
                if order.convert_to_market_after_expiry:
                    order.order_type = "MARKET"
                    order.modify(self.broker)
                elif order.cancel_after_expiry:
                    order.cancel(broker=self.broker)

    @property
    def completed_orders(self) -> List[Order]:
        return [order for order in self.orders if order.is_complete]

    @property
    def pending_orders(self) -> List[Order]:
        return [order for order in self.orders if order.is_pending]

    def add(self, order: Order) -> Optional[str]:
        """
        Add an order to the existing compound order
        """
        order.parent_id = self.id
        if not (order.connection):
            order.connection = self.connection
        order.save_to_db()
        self.orders.append(order)
        return order.id

    def save(self):
        """
        Save all orders to database
        """
        if self.count > 0:
            for order in self.orders:
                order.save_to_db()
        pass



from omspy.models import Quote
from omspy.utils import tick
from typing import List
from pydantic import BaseModel


class MarketDepth(BaseModel):
    bids: List[Quote]
    asks: List[Quote]
    tick: float = 0.05

    @property
    def midpoint(self) -> float:
        a, b = self.bids[0].price, self.asks[0].price
        mp = abs(b - a) / 2
        tck = tick(min(a, b) + mp, 0.15)
        return round(tick(min(a, b) + mp, tick_size=self.tick), 2)

    def bid(self, n: int = 0) -> float:
        """
        Get the bid price at the specified depth
        """
        return self.bids[n].price

    def ask(self, n: int = 0) -> float:
        """
        Get the ask price at the specified depth
        """
        return self.asks[n].price

    def sort(self):
        """
        Sort asks and bids based on price
        sorting is done in place replacing asks and bids
        """
        self.bids.sort(key=lambda x: x.price, reverse=True)
        self.asks.sort(key=lambda x: x.price)
